"""
File containing all of the run tags in the mlflow. namespace.
"""
MLFLOW_DATABRICKS_NOTEBOOK_ID = "mlflow.databricks.notebookID"
MLFLOW_DATABRICKS_NOTEBOOK_PATH = "mlflow.databricks.notebookPath"
MLFLOW_DATABRICKS_WEBAPP_URL = "mlflow.databricks.webappURL"
MLFLOW_DATABRICKS_RUN_URL = "mlflow.databricks.runURL"
MLFLOW_DATABRICKS_SHELL_JOB_ID = "mlflow.databricks.shellJobID"
MLFLOW_DATABRICKS_SHELL_JOB_RUN_ID = "mlflow.databricks.shellJobRunID"
MLFLOW_RUN_NAME = "mlflow.runName"
MLFLOW_GIT_BRANCH_NAME = "mlflow.gitBranchName"
MLFLOW_GIT_REPO_URL = "mlflow.gitRepoURL"
MLFLOW_PARENT_RUN_ID = "mlflow.parentRunId"
